//---------------------------------------------------------------------------
#ifndef FileH
#define FileH
//---------------------------------------------------------------------------
#include <string>
#include <fstream>
#include <algorithm>
#include <gimite/string_algo.h>
#include <gimite/temp_array.h>
#include "Maprof.h"
#include "Definitions.h"
#include "StrProcess.h"
//---------------------------------------------------------------------------
enum TReturnCode{//sR[h
	rcWin,//Windows(\r\n)
  rcUnix,//Unix(\n)
  rcMac//Mac(\r)
};

enum TFileOperationResult{//t@Č
	forOk,//ُ햳
  forError,//G[
  forBinary,//oCit@C
  forCharCodeMismatch,//R[hsv
  forChanged,//ɕύXꂽ
  forSizeOver,//t@CTCY𒴂Ă
};

enum TCharCode{//R[h(*1)
	ccAuto,//
	ccSjis,//Shift-JIS
	ccUtf16,//UTF-16(Unicode)
	ccAscii,//Ascii
	ccEuc,//EUC
	ccJis,//JIS
	ccUtf8,//UTF-8
	ccUtf8N,//UTF-8NiBOMzj
	ccUnknown,//s
};

//*1	frmOptioñR{{bNX̃Xg
//    ԂɏƁB
//---------------------------------------------------------------------------
std::istream& operator>>(std::istream& is, TCharCode& ccode);
std::ostream& operator<<(std::ostream& os, TCharCode ccode);
//---------------------------------------------------------------------------
struct TFileOperationInfo{//t@C̏
	bool bCheckCharCode_;//R[h`FbN
  int sizeMaxFile_;//t@CTCY̏ioCgj
	TFileOperationInfo(): bCheckCharCode_(false), sizeMaxFile_(INT_MAX){}
};

//---------------------------------------------------------------------------
//t@C̊NX
template <class TChar>
class TFileBase
{
public:
	typedef std::basic_string<TChar> TString;
  typedef TFileOperationResult TResult;
	typedef TFileOperationInfo TInfo;

	virtual ~TFileBase(){}
  void SetPath(const std::string& path){path_= path;}
  TString& Text(){return text_;}//eLXg

protected:
	TResult LoadTextSimply(std::string& text, const TInfo& info= TInfo());
  	//P1oCgPʃeLXg[h(*1)
	TResult SaveTextSimply(const std::string& text);
		//P1oCgPʃeLXgۑ(*1)
  int FileSize(customifstream& fs);
    //t@CTCY擾B
	void DecodeReturnCode(TString& text);
		//݂̉sR[hۑUNIXR[hɕϊ
	void EncodeReturnCode(TString& text);
		//sR[hɖ߂
	bool IsWideCharCode(TCharCode ccode);
		//ChR[h

private:
	static const std::size_t sizeBuf_= 1024*1024;
		//ǂݍ݃obt@TCYF1MB

	std::string path_;//pX
	TReturnCode rcode_;//sR[h
	TString text_;//eLXg

};

//*1	R[hϊsR[hϊsȂ
//---------------------------------------------------------------------------
template <class TChar>
class TFile;
//---------------------------------------------------------------------------
//VOoCgt@C
template <>
class TFile<char>: public TFileBase<char>
{
public:
	typedef TFileBase<char>::TString TString;
	typedef TFileBase<char>::TResult TResult;
	typedef TFileOperationInfo TInfo;

	TResult LoadText(const TInfo& info= TInfo());//eLXg[h
	TResult SaveText(const TInfo& info= TInfo());//eLXgZ[u
	void SetCharCode(TCharCode cc){assert(cc==ccAscii);}
		//R[h̐ݒ

};

//---------------------------------------------------------------------------
//}`oCgt@C
template <>
class TFile<wchar_t>: public TFileBase<wchar_t>
{
public:
	typedef TFileBase<wchar_t>::TString TString;
	typedef TFileBase<wchar_t>::TResult TResult;
  typedef TFileOperationInfo TInfo;

  TResult LoadText(const TInfo& info= TInfo());//eLXg[h
  TResult SaveText(const TInfo& info= TInfo());//eLXgZ[u
	void SetCharCode(TCharCode cc){ccode_= cc;}
		//R[h̐ݒ

private:
	TCharCode ccode_;//R[h

	TCharCode AnalyzeCharCode(const std::string& text);
		//R[h𔻒
	bool DecodeCharCode(const std::string& sSrc, TString& tsDest);
		//R[hUnicodeɕϊ(*1)
	void EncodeCharCode(const TString& tsSrc, std::string& sDest);
		//R[hɖ߂

};

//*1	R[hႤۂfalse
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//ev[g֐̎
//---------------------------------------------------------------------------
template <class T>
TFileBase<T>::TResult TFileBase<T>::LoadTextSimply(
    std::string& text, const TInfo& info/*= TInfo()*/)
  //P1oCgPʃeLXg[h(*1)
{
	GIMITE_MAPROFILED("TFileBase<T>::LoadTextSimply@1");
	text.clear();
	customifstream fs(sizeBuf_, path_.c_str(),
		std::ios::in|std::ios::binary);
	if (!fs) return forError;
  if (FileSize(fs) > info.sizeMaxFile_) return forSizeOver;
	while (fs){
		std::size_t sz= fs.rdbuf()->in_avail();
		if (sz==std::size_t(-1) || sz==0) sz= 1;
		gimite::temp_array<char> buf(sz);
		if (fs.read(buf.get(), sz))
			text.append(buf.get(), sz);
	}
	return forOk;
}

//---------------------------------------------------------------------------
template <class T>
TFileBase<T>::TResult TFileBase<T>::SaveTextSimply(const std::string& text)
	//P1oCgPʃeLXgۑ(*1)
{
	std::ofstream fs(path_.c_str(), std::ios::out|std::ios::binary);
	if (!fs) return forError;
	if (!(fs<<text)) return forError;
	return forOk;
}

//---------------------------------------------------------------------------
template <class T>
int TFileBase<T>::FileSize(customifstream& fs)
  //t@CTCY擾
{
  //Xg[Jł邱Ƃ肵ĂB
  //܂A2GBȏ̃t@Cɑ΂Ă͐lԂȂB
  int size= fs.rdbuf()->pubseekoff(0, ios_base::end);
  fs.rdbuf()->pubseekoff(0, ios_base::beg);
  return size;
}

//---------------------------------------------------------------------------
template <class T>
void TFileBase<T>::DecodeReturnCode(TString& text)
	//݂̉sR[hۑUNIXR[hɕϊ
{
	GIMITE_MAPROFILED("TFileBase<T>::DecodeReturnCode");
	//sR[h
	TString::size_type pos1;
	for (pos1= 0; pos1<text.length(); ++pos1){
		if (text[pos1]==T('\n') || text[pos1]==T('\r')) break;
	}
	if (pos1==text.length()){
		rcode_= rcUnix;//sȂUnixƂ݂Ȃ
	}else if (text[pos1]==T('\n')){
		rcode_= rcUnix;
	}else if (pos1+1<text.length() && text[pos1+1]==T('\n')){
		rcode_= rcWin;
	}else{
		rcode_= rcMac;
	}

	//sR[h
	if (rcode_==rcWin){
		//{glbN̂ōȕ@
		std::size_t lenMove= 0;
		for (std::size_t pos= 0; pos<text.length(); ++pos){
      assert(pos>=lenMove);
			if (text[pos]==T('\r')) ++lenMove;
			else text[pos-lenMove]= text[pos];
		}
		text.resize(text.length()-lenMove);
	}else if (rcode_==rcMac){
		std::replace(text.begin(), text.end(), T('\r'), T('\n'));
	}
}

//---------------------------------------------------------------------------
template <class T>
void TFileBase<T>::EncodeReturnCode(TString& text)
	//sR[hɖ߂
{
	GIMITE_MAPROFILED("TFileBase<T>::EncodeReturnCode");
	if (rcode_==rcWin){
		text= gimite::str_replace_all_copy<T>(text,
      ToTstr<T>("\n"), ToTstr<T>("\r\n"));
	}else if (rcode_==rcMac){
		std::replace(text.begin(), text.end(), T('\n'), T('\r'));
	}
}

//---------------------------------------------------------------------------
template <class T>
bool TFileBase<T>::IsWideCharCode(TCharCode ccode)
	//ChR[h
{
	switch (ccode){
		case ccUtf16: return true;
		case ccSjis: case ccAscii: case ccEuc: case ccJis: case ccUtf8:
			case ccUtf8N: return false;
		default: assert(false); return false;
	}
}

//---------------------------------------------------------------------------
#endif
