//---------------------------------------------------------------------------

#pragma hdrstop

#include <boost/scoped_array.hpp>
#include <fstream>
#include <algorithm>
#include <iomanip>
#include <cstring>
#include "GTS.h"
using namespace std;
using namespace boost;
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
namespace gimite{
//---------------------------------------------------------------------------
//t@Cm[h쐬
bool TGTSNode::LoadFromFile(const char* path)
{
	ifstream fs(path);
	if (!fs) return false;
	return ReadFromStream(fs);
}

//---------------------------------------------------------------------------
//t@CɃm[hۑ
bool TGTSNode::SaveToFile(const char* path)const
{
	ofstream fs(path);
	if (!fs) return false;
	return WriteToStream(fs, 0);
}

//---------------------------------------------------------------------------
//IuWFNgǂݍ
bool TGTSNode::ReadObject(const std::string& sKey, TGTSObject* obj)
{
	shared_ptr<TGTSNode> node= PopChildNode(sKey);
	if (!node.get()) return false;
	obj->ReadMembers(node.get());
	return true;
}

//---------------------------------------------------------------------------
//IuWFNg
void TGTSNode::WriteObject(const std::string& sKey, const TGTSObject* obj)
{
	TGTSNode* node= AddChildNode(sKey);
	obj->WriteMembers(node);
}

//---------------------------------------------------------------------------
//oCif[^̓ǂݍ
bool TGTSNode::ReadBinary(const std::string& sKey, void* buffer, int size)
{
	string s;
	if (!ReadString(sKey, s)) return false;
	istringstream ss(s);
	int n;
	scoped_array<unsigned char> bufferT(new unsigned char[size]);
	for (int i= 0; i<size; ++i){
		ss>>hex>>n;
		bufferT[i]= static_cast<unsigned char>(n);
	}
	if (ss) memcpy(buffer, bufferT.get(), size);
	return ss;
}

//---------------------------------------------------------------------------
//oCif[^̓ǂݍ
void TGTSNode::WriteBinary(const std::string& sKey, const void* buffer, int size)
{
	ostringstream ss;
	for (int i= 0; i<size; ++i){
		int n= *(static_cast<const unsigned char*>(buffer)+i);
		ss<<hex<<setw(2)<<setfill('0')<<n<<' ';
	}
	WriteString(sKey, ss.str());
}

//---------------------------------------------------------------------------
//܂ǂłȂf[^̐
int TGTSNode::DataCount(const std::string& sKey)const
{
	return count(data_.begin(), data_.end(), TDatum(sKey, ""));
}

//---------------------------------------------------------------------------
//܂ǂłȂIuWFNg̐
int TGTSNode::ObjectsCount(const std::string& sKey)const
{
	return count(infosChildNode_.begin(), infosChildNode_.end(),
		TNodeInfo(sKey, TNodePtr()));
}

//---------------------------------------------------------------------------
//qm[h̎o
TGTSNode::TNodePtr TGTSNode::PopChildNode(const std::string& sKey)
{
	TNodeInfos::iterator ndit
		= find(infosChildNode_.begin(), infosChildNode_.end(),
		TNodeInfo(sKey, shared_ptr<TGTSNode>()));
	if (ndit!=infosChildNode_.end()){
		shared_ptr<TGTSNode> node= ndit->second;
		infosChildNode_.erase(ndit);
		return node;
	}else{
		return shared_ptr<TGTSNode>();
	}
}

//---------------------------------------------------------------------------
//qm[h̒ǉ
TGTSNode* TGTSNode::AddChildNode(const std::string& sKey)
{
	TNodeInfo info(sKey, shared_ptr<TGTSNode>(new TGTSNode));
	infosChildNode_.push_back(info);
	return info.second.get();
}

//---------------------------------------------------------------------------
//m[hǂݍ
bool TGTSNode::ReadFromStream(std::istream& is)
{
	typedef string::size_type TPos;
	data_.clear();
	infosChildNode_.clear();
	string sLine;
	while (getline(is, sLine, '\n')){
		TPos pos= 0;
		TPos len= sLine.length();
		while (pos<len && isspace(sLine[pos])) ++pos;
		if (pos<len && sLine[pos]=='}') break;
		TPos posBg= pos;
		while (pos<len && sLine[pos]!='=' && sLine[pos]!='{') ++pos;
		if (pos>=len) continue;
		string sKey= sLine.substr(posBg, pos-posBg);
		if (sLine[pos]=='{'){
			TGTSNode* node= AddChildNode(sKey);
			node->ReadFromStream(is);
		}else{
			data_.push_back(TDatum(sKey, sLine.substr(pos+1)));
		}
	}
	return true;
}

//---------------------------------------------------------------------------
//m[h
bool TGTSNode::WriteToStream(std::ostream& os, int nIndent)const
{
	for (TData::const_iterator dtit= data_.begin(); dtit!=data_.end(); ++dtit){
		os<<string(nIndent*2, ' ');
		os<<(dtit->first)<<'='<<(dtit->second)<<endl;
	}
	for (TNodeInfos::const_iterator ifit= infosChildNode_.begin();
			ifit!=infosChildNode_.end(); ++ifit){
		os<<string(nIndent*2, ' ');
		os<<(ifit->first)<<'{'<<endl;
		ifit->second->WriteToStream(os, nIndent+1);
		os<<string(nIndent*2, ' ')<<'}'<<endl;
	}
	return true;
}

//---------------------------------------------------------------------------
//f[^̓ǂݍ
bool TGTSNode::ReadString(const std::string& sKey, std::string& s)
{
	TData::iterator dtit= find(data_.begin(), data_.end(), TDatum(sKey, ""));
	if (dtit!=data_.end()){
		s= dtit->second;
		data_.erase(dtit);
		return true;
	}else{
		return false;
	}
}

//---------------------------------------------------------------------------
//f[^̏
void TGTSNode::WriteString(const std::string& sKey, const std::string& s)
{
	data_.push_back(TDatum(sKey, s));
}

//---------------------------------------------------------------------------
//t@C烍[h
bool TGTSObject::LoadFromFile(const char* path)
{
	TGTSNode node;
	if (!node.LoadFromFile(path)) return false;
	ReadMembers(&node);
	return true;
}

//---------------------------------------------------------------------------
//t@Cɕۑ
bool TGTSObject::SaveToFile(const char* path)const
{
	TGTSNode node;
	WriteMembers(&node);
	return node.SaveToFile(path);
}

//---------------------------------------------------------------------------
}//namespace gimite
//---------------------------------------------------------------------------

