//---------------------------------------------------------------------------
#ifndef GIMITE_GTSH
#define GIMITE_GTSH
//---------------------------------------------------------------------------
#include <sstream>
#include <string>
#include <list>
#include <iostream>
#include <boost/config.hpp>
#include <boost/shared_ptr.hpp>
//---------------------------------------------------------------------------
namespace gimite{
//---------------------------------------------------------------------------
class TGTSObject;
//---------------------------------------------------------------------------
//ƐFXȌ^̑ݕϊ`邽߂̃NX
class TGTSString
{
public:
	TGTSString(const std::string& s= ""): s_(s){}
	std::string Str()const{return s_;}
	TGTSString& operator=(const std::string& s){s_= s; return *this;}

private:
	std::string s_;

};

template <class T>
inline bool operator>>(const TGTSString& gs, T& v)
{
	std::istringstream ss(gs.Str());
	return static_cast<std::istream&>(ss)>>v;
}
template <class T>
inline void operator<<(TGTSString& gs, const T& v)
{
	std::ostringstream ss;
	static_cast<std::ostream&>(ss)<<v;
	gs= ss.str();
}

#ifdef BOOST_NO_TEMPLATE_PARTIAL_SPECIALIZATION
inline bool operator>>(const TGTSString& gs, std::string& v)
{
	v= gs.Str();
	return true;
}
inline void operator<<(TGTSString& gs, const std::string& v){gs= v;}
#else
template <class T, class A>
inline bool operator>>(const TGTSString& gs,
	std::basic_string<char, T, A>& v)
{
	v= gs.Str();
	return true;
}
template <class T, class A>
inline void operator<<(TGTSString& gs,
	const std::basic_string<char, T, A>& v){gs= v;}
#endif

//---------------------------------------------------------------------------
//IuWFNgɑΉm[h
class TGTSNode
{
	typedef boost::shared_ptr<TGTSNode> TNodePtr;

public:
	//f[^̓ǂݍ
	template <class T>
	bool Read(const std::string& sKey, T& v){
		std::string s;
		if (!ReadString(sKey, s)) return false;
		TGTSString gs(s);
		return gs>>v;
	}
	//f[^̏
	template <class T>
	void Write(const std::string& sKey, const T& v){
		TGTSString gs;
		gs<<v;
		WriteString(sKey, gs.Str());
	}
	//IuWFNgǂݍ
	bool ReadObject(const std::string& sKey, TGTSObject* obj);
	//IuWFNg
	void WriteObject(const std::string& sKey, const TGTSObject* obj);
	//oCif[^̓ǂݍ
	bool ReadBinary(const std::string& sKey, void* buffer, int size);
	//oCif[^̓ǂݍ
	void WriteBinary(const std::string& sKey, const void* buffer, int size);
	//܂ǂłȂf[^̐
	int DataCount(const std::string& sKey)const;
	//܂ǂłȂIuWFNg̐
	int ObjectsCount(const std::string& sKey)const;

	//t@Cm[h쐬
	bool LoadFromFile(const char* path);
	//t@CɃm[hۑ
	bool SaveToFile(const char* path)const;

private:
	struct TDatum{//L[ƒl̑g
		std::string first, second;
		TDatum(const std::string& s1= "", const std::string& s2= "")
			: first(s1), second(s2){}
		bool operator==(const TDatum& dt2)const
			{return first==dt2.first;}
		bool operator!=(const TDatum& dt2)const
			{return !(*this==dt2);}

	};
	typedef std::list<TDatum> TData;
	struct TNodeInfo{
		std::string first;
		TNodePtr second;
		TNodeInfo(const std::string& s= "", TNodePtr node= TNodePtr())
			: first(s), second(node){}
		bool operator==(const TNodeInfo& info2)const
			{return first==info2.first;}
		bool operator!=(const TNodeInfo& info2)const
			{return !(*this==info2);}
	};
	typedef std::list<TNodeInfo> TNodeInfos;

	TData data_;//m[hɊ܂܂f[^
	TNodeInfos infosChildNode_;//qm[h

	//m[hǂݍ
	bool ReadFromStream(std::istream& is);
	//m[h
	bool WriteToStream(std::ostream& os, int nIndent)const;
	//f[^̓ǂݍ
	bool ReadString(const std::string& sKey, std::string& s);
	//f[^̏
	void WriteString(const std::string& sKey, const std::string& s);
	//qm[h̎o
	TNodePtr PopChildNode(const std::string& sKey);
	//qm[h̒ǉ
	TGTSNode* AddChildNode(const std::string& sKey);

};

//---------------------------------------------------------------------------
//ۑΏۂ̃IuWFNg̊NX
class TGTSObject
{
public:
	virtual ~TGTSObject(){}
	//t@C烍[h
	bool LoadFromFile(const char* path);
	//t@Cɕۑ
	bool SaveToFile(const char* path)const;

protected:
	friend class TGTSNode;

	//oǂݍ
	virtual void ReadMembers(TGTSNode* node)= 0;
	//o
	virtual void WriteMembers(TGTSNode* node)const= 0;

};

//---------------------------------------------------------------------------
}//namespace gimite
//---------------------------------------------------------------------------
#endif
