//---------------------------------------------------------------------------
/*
     Xbh[Jϐw|C^ thread_local_ptr 1.0  by Gimite

*/
//---------------------------------------------------------------------------
#ifndef GIMITE_thread_local_ptrH
#define GIMITE_thread_local_ptrH
//---------------------------------------------------------------------------
#include <vector>
#include <cassert>
#include <algorithm>
#include <boost/utility.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/thread/tss.hpp>
#include <gimite/singleton_ptr.h>
//---------------------------------------------------------------------------
#if defined(__BORLANDC__) && __BORLANDC__>=0x560 && __BORLANDC__<0x570 \
  && (defined(__SGI_STL_PORT) || defined(_STLPORT_VERSION))

namespace std{
  using _STL::remove;
}

#endif
//---------------------------------------------------------------------------
namespace gimite{
//---------------------------------------------------------------------------
template <class T>
void deleter(void* p){
  delete static_cast<T*>(p);
}

//---------------------------------------------------------------------------
//XbhƂ1pӂ̈
class thread_local_area_t
{
public:
  typedef std::vector<void*> vptrs_t;
  typedef std::vector<void (*)(void*)> deleters_t;
  typedef vptrs_t::size_type vsize_t;
  typedef boost::mutex::scoped_lock lock_t;

private:
  vptrs_t vars_;
  deleters_t deleters_;
  boost::mutex mtx_;

  void resize(vsize_t size){
    lock_t lck(mtx_);
    vars_.resize(size);
    deleters_.resize(size);
  }

public:
  thread_local_area_t();
  ~thread_local_area_t();
  void*& get(vsize_t n){
    if (n>=vars_.size()) resize(n+1);
    return vars_[n];
  }
  void set_deleter(vsize_t n, void (*deleter)(void*)){
    if (n>=vars_.size()) resize(n+1);
    deleters_[n]= deleter;
  }
  void delete_var(vsize_t n){//(*1)
    lock_t lck(mtx_);
    assert(vars_.size()==deleters_.size());
    if (n<vars_.size() && deleters_[n]) deleters_[n](vars_[n]);
    if (n<vars_.size()) vars_[n]= 0;
  }

};

//*1  XbhĂяoB
//---------------------------------------------------------------------------
//Xbh[Jϐ̊Ǘ
class thread_local_manager_t: boost::noncopyable
{
public:
  typedef std::vector<thread_local_area_t*> areas_t;
  typedef areas_t::size_type asize_t;
  typedef boost::mutex::scoped_lock lock_t;
  typedef thread_local_area_t::vsize_t vsize_t;
  friend class gimite::singleton_ptr<thread_local_manager_t>;

private:
  areas_t areas_;
  boost::thread_specific_ptr<thread_local_area_t> area_;
  std::vector<bool> is_used_;
  boost::mutex mtx_;

  thread_local_manager_t(){}
  ~thread_local_manager_t(){
    area_.reset(0);
      //CXbharea̔jthread_specific_ptrɔC
      //BCC 5.6.4ŎsG[ɂȂ
  }

public:
  vsize_t regist(){
    lock_t lck(mtx_);
    vsize_t i;
    for (i= 0; i<is_used_.size(); ++i)
      if (!is_used_[i]) break;
    if (i>=is_used_.size())
      is_used_.push_back(true);
    else
      is_used_[i]= true;
    return i;
  }
  void unregist(vsize_t n){
    lock_t lck(mtx_);
    assert(n<is_used_.size());
    is_used_[n]= false;
    for (asize_t i= 0; i<areas_.size(); ++i)
      areas_[i]->delete_var(n);
  }
  void add_area(thread_local_area_t* area){
    lock_t lck(mtx_);
    areas_.push_back(area);
  }
  void erase_area(thread_local_area_t* area){
    lock_t lck(mtx_);
    areas_.erase(std::remove(areas_.begin(), areas_.end(), area), areas_.end());
  }
  thread_local_area_t* area(){
    if (!area_.get())
      area_.reset(new thread_local_area_t);
    return area_.get();
  }

};

static singleton_ptr<thread_local_manager_t> thread_local_manager;
//---------------------------------------------------------------------------
//Xbh[Jϐw|C^
template <class T>
class thread_local_ptr
{
  typedef thread_local_manager_t::vsize_t vsize_t;
  
  vsize_t n_;
  
  static thread_local_manager_t* manager(){
    return thread_local_manager.get();
  }
  static thread_local_area_t* area(){
    return manager()->area();
  }

public:
  thread_local_ptr(): n_(manager()->regist()){}
  ~thread_local_ptr(){
    manager()->unregist(n_);
  }
  T* get()const{ return static_cast<T*>(area()->get(n_)); }
  T& operator*()const{ return *get(); }
  T* operator->()const{ return get(); }
  void reset(T* p= 0){
    thread_local_area_t* ar= area();
    ar->get(n_)= p;
    ar->set_deleter(n_, &deleter<T>);
  }
  T* release(){
    thread_local_area_t* ar= area();
    T* p= static_cast<T*>(ar->get(n_));
    ar->get(n_)= 0;
    return p;
  }
  
};

//---------------------------------------------------------------------------
//XbhƂ1݂CX^Xw|C^
template <class T>
class thread_singleton_ptr
{
  singleton_ptr<thread_local_ptr<T> > ptr_;
  
public:
  T* get()const{ return ptr_->get(); }
  T& operator*()const{ return *get(); }
  T* operator->()const{ return get(); }
  void reset(T* p= 0){ ptr_->reset(p); }
  T* release(){ return ptr_->release(); }
  
};

//---------------------------------------------------------------------------
inline thread_local_area_t::thread_local_area_t(){
  thread_local_manager->add_area(this);
}

inline thread_local_area_t::~thread_local_area_t(){
  assert(thread_local_manager.get());
  thread_local_manager->erase_area(this);
  for (vsize_t i= 0; i<vars_.size(); ++i)
    delete_var(i);
}

//---------------------------------------------------------------------------
}//namespace gimite
//---------------------------------------------------------------------------
#endif
